"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ASTProvider = void 0;
const fs_1 = require("fs");
const tsyringe_1 = require("tsyringe");
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_uri_1 = require("vscode-uri");
const elmWorkspaceMatcher_1 = require("../util/elmWorkspaceMatcher");
const textDocumentEvents_1 = require("../util/textDocumentEvents");
const treeUtils_1 = require("../util/treeUtils");
const utils_1 = require("../util/utils");
class ASTProvider {
    constructor() {
        this.treeChangeEvent = new vscode_languageserver_1.Emitter();
        this.onTreeChange = this.treeChangeEvent.event;
        this.pendingRenames = new Map();
        this.handleChangeTextDocument = (params) => {
            var _a, _b, _c, _d, _e;
            this.connection.console.info(`Changed text document, going to parse it. ${params.textDocument.uri}`);
            const forest = params.program.getForest(false); // Don't synchronize the forest, we are only looking at the tree
            const document = params.textDocument;
            // Source file could be undefined here
            let tree = (_a = params.sourceFile) === null || _a === void 0 ? void 0 : _a.tree;
            let hasContentChanges = false;
            if ("contentChanges" in params) {
                hasContentChanges = true;
                for (const change of params.contentChanges) {
                    if ("range" in change) {
                        tree === null || tree === void 0 ? void 0 : tree.edit(this.getEditFromChange(change, tree.rootNode.text));
                    }
                }
            }
            const pendingRenameUri = this.pendingRenames.get(document.uri);
            this.pendingRenames.delete(document.uri);
            // Remove the old tree
            if (pendingRenameUri) {
                forest.removeTree(document.uri);
            }
            const newText = (_c = (_b = this.documentEvents.get(document.uri)) === null || _b === void 0 ? void 0 : _b.getText()) !== null && _c !== void 0 ? _c : fs_1.readFileSync(vscode_uri_1.URI.parse(document.uri).fsPath, "utf8");
            const newTree = this.parser.parse(newText, hasContentChanges ? tree : undefined);
            let changedDeclaration;
            tree === null || tree === void 0 ? void 0 : tree.getChangedRanges(newTree).map((range) => [
                tree === null || tree === void 0 ? void 0 : tree.rootNode.descendantForPosition(range.startPosition),
                tree === null || tree === void 0 ? void 0 : tree.rootNode.descendantForPosition(range.endPosition),
            ]).map(([startNode, endNode]) => [
                startNode
                    ? treeUtils_1.TreeUtils.findParentOfType("value_declaration", startNode, true)
                    : undefined,
                endNode
                    ? treeUtils_1.TreeUtils.findParentOfType("value_declaration", endNode, true)
                    : undefined,
            ]).forEach(([startNode, endNode]) => {
                if (startNode &&
                    endNode &&
                    startNode.id === endNode.id &&
                    treeUtils_1.TreeUtils.getTypeAnnotation(startNode)) {
                    changedDeclaration = startNode;
                    params.program.getTypeCache().invalidateValueDeclaration(startNode);
                }
            });
            if (!changedDeclaration) {
                params.program.getTypeCache().invalidateProject();
            }
            tree = newTree;
            if (tree) {
                // Reuse old source file for most cases
                const isTestFile = params.sourceFile
                    ? params.sourceFile.isTestFile
                    : (_e = (_d = params.program
                        .getSourceDirectoryOfFile(document.uri)) === null || _d === void 0 ? void 0 : _d.endsWith("tests")) !== null && _e !== void 0 ? _e : false;
                const treeContainer = forest.setTree(pendingRenameUri !== null && pendingRenameUri !== void 0 ? pendingRenameUri : document.uri, true, true, tree, isTestFile);
                // The workspace now needs to be synchronized
                params.program.markAsDirty();
                setImmediate(() => {
                    if (tree) {
                        this.treeChangeEvent.fire({
                            treeContainer,
                            declaration: changedDeclaration,
                        });
                    }
                });
            }
        };
        this.parser = tsyringe_1.container.resolve("Parser");
        this.connection = tsyringe_1.container.resolve("Connection");
        this.documentEvents = tsyringe_1.container.resolve(textDocumentEvents_1.TextDocumentEvents);
        this.documentEvents.on("change", new elmWorkspaceMatcher_1.ElmWorkspaceMatcher((params) => vscode_uri_1.URI.parse(params.textDocument.uri)).handle(this.handleChangeTextDocument.bind(this)));
        this.documentEvents.on("open", new elmWorkspaceMatcher_1.ElmWorkspaceMatcher((params) => vscode_uri_1.URI.parse(params.textDocument.uri)).handle(this.handleChangeTextDocument.bind(this)));
    }
    addPendingRename(oldUri, newUri) {
        this.pendingRenames.set(oldUri, newUri);
    }
    getEditFromChange(change, text) {
        const [startIndex, endIndex] = utils_1.Utils.getIndicesFromRange(change.range, text);
        return {
            startIndex,
            oldEndIndex: endIndex,
            newEndIndex: startIndex + change.text.length,
            startPosition: this.toTSPoint(change.range.start),
            oldEndPosition: this.toTSPoint(change.range.end),
            newEndPosition: this.toTSPoint(this.addPositions(change.range.start, this.textToPosition(change.text))),
        };
    }
    textToPosition(text) {
        const lines = text.split(/\r\n|\r|\n/);
        return {
            line: lines.length - 1,
            character: lines[lines.length - 1].length,
        };
    }
    addPositions(pos1, pos2) {
        return {
            line: pos1.line + pos2.line,
            character: pos1.character + pos2.character,
        };
    }
    toTSPoint(position) {
        return { row: position.line, column: position.character };
    }
}
exports.ASTProvider = ASTProvider;
//# sourceMappingURL=astProvider.js.map