"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CodeActionProvider = void 0;
const fs_1 = require("fs");
const tsyringe_1 = require("tsyringe");
const util_1 = require("util");
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_uri_1 = require("vscode-uri");
const elmPackageCache_1 = require("../elmPackageCache");
const elmWorkspaceMatcher_1 = require("../util/elmWorkspaceMatcher");
const multiMap_1 = require("../util/multiMap");
const treeUtils_1 = require("../util/treeUtils");
const diagnosticsProvider_1 = require("./diagnostics/diagnosticsProvider");
const elmLsDiagnostics_1 = require("./diagnostics/elmLsDiagnostics");
const elmMakeDiagnostics_1 = require("./diagnostics/elmMakeDiagnostics");
const fileDiagnostics_1 = require("./diagnostics/fileDiagnostics");
const exposeUnexposeHandler_1 = require("./handlers/exposeUnexposeHandler");
const moveRefactoringHandler_1 = require("./handlers/moveRefactoringHandler");
class CodeActionProvider {
    constructor() {
        var _a;
        this.settings = tsyringe_1.container.resolve("Settings");
        this.elmMake = tsyringe_1.container.resolve(elmMakeDiagnostics_1.ElmMakeDiagnostics);
        this.elmDiagnostics = tsyringe_1.container.resolve(elmLsDiagnostics_1.ElmLsDiagnostics);
        this.connection = tsyringe_1.container.resolve("Connection");
        this.diagnosticsProvider = tsyringe_1.container.resolve(diagnosticsProvider_1.DiagnosticsProvider);
        this.onCodeAction = this.onCodeAction.bind(this);
        this.connection.onCodeAction(this.diagnosticsProvider.interruptDiagnostics(() => new elmWorkspaceMatcher_1.ElmWorkspaceMatcher((param) => vscode_uri_1.URI.parse(param.textDocument.uri)).handle(this.onCodeAction.bind(this))));
        this.connection.onRequest(vscode_languageserver_1.CodeActionResolveRequest.method, new elmWorkspaceMatcher_1.ElmWorkspaceMatcher((codeAction) => vscode_uri_1.URI.parse(codeAction.data.uri)).handleResolve((codeAction, program, sourceFile) => this.onCodeActionResolve(codeAction, program, sourceFile)));
        if ((_a = this.settings.extendedCapabilities) === null || _a === void 0 ? void 0 : _a.moveFunctionRefactoringSupport) {
            new moveRefactoringHandler_1.MoveRefactoringHandler();
        }
        new exposeUnexposeHandler_1.ExposeUnexposeHandler();
        setTimeout(() => {
            void new elmPackageCache_1.ElmPackageCache((path) => __awaiter(this, void 0, void 0, function* () {
                return JSON.parse(yield util_1.promisify(fs_1.readFile)(path, { encoding: "utf-8" }));
            })).loadAllPackageModules();
        }, 5000);
    }
    static registerCodeAction(registration) {
        registration.errorCodes.forEach((code) => {
            CodeActionProvider.errorCodeToRegistrationMap.set(code, registration);
        });
    }
    static registerRefactorAction(name, registration) {
        this.refactorRegistrations.set(name, registration);
    }
    static getDiagnostics(params) {
        return [
            ...params.program.getSyntacticDiagnostics(params.sourceFile),
            ...params.program.getSemanticDiagnostics(params.sourceFile),
            ...params.program.getSuggestionDiagnostics(params.sourceFile),
        ];
    }
    static forEachDiagnostic(params, errorCodes, callback) {
        CodeActionProvider.getDiagnostics(params).forEach((diagnostic) => {
            if (typeof diagnostic.code === "string" &&
                errorCodes.includes(diagnostic.code)) {
                callback(diagnostic);
            }
        });
    }
    static getCodeAction(params, title, edits, command) {
        const changes = { [params.sourceFile.uri]: edits };
        return {
            title,
            kind: vscode_languageserver_1.CodeActionKind.QuickFix,
            edit: { changes },
            isPreferred: true,
            command,
        };
    }
    static getFixAllCodeAction(title, params, errorCodes, fixId, callback) {
        const edits = [];
        const changes = {
            [params.sourceFile.uri]: edits,
        };
        const diagnostics = [];
        CodeActionProvider.forEachDiagnostic(params, errorCodes, (diagnostic) => {
            diagnostics.push(diagnostic);
            callback(edits, diagnostic);
        });
        return {
            title,
            kind: vscode_languageserver_1.CodeActionKind.QuickFix,
            diagnostics,
            edit: { changes },
            data: fixId,
        };
    }
    onCodeAction(params) {
        this.connection.console.info("A code action was requested");
        const make = this.elmMake.onCodeAction(params);
        const elmDiagnostics = this.elmDiagnostics.onCodeAction(params);
        const results = [];
        // For each diagnostic in the context, get the code action registration that
        // handles the diagnostic error code and ask for the code actions for that error
        // and the fix all code action for that error if there are other diagnostics with
        // the same error code
        params.context.diagnostics.forEach((diagnostic) => {
            const registrations = CodeActionProvider.errorCodeToRegistrationMap.getAll(diagnostic.data.code);
            // Set the params range to the diagnostic range so we get the correct nodes
            params.range = diagnostic.range;
            results.push(...treeUtils_1.flatMap(registrations, (reg) => {
                var _a, _b;
                const codeActions = (_b = (_a = reg
                    .getCodeActions(params)) === null || _a === void 0 ? void 0 : _a.map((codeAction) => this.addDiagnosticToCodeAction(codeAction, diagnostic))) !== null && _b !== void 0 ? _b : [];
                if (codeActions.length > 0 &&
                    !results.some(
                    // Check if there is already a "fix all" code action for this fix
                    (codeAction) => /* fixId */ codeAction.data === reg.fixId) &&
                    CodeActionProvider.getDiagnostics(params).some((diag) => !fileDiagnostics_1.diagnosticsEquals(diagnosticsProvider_1.convertFromAnalyzerDiagnostic(diag), diagnostic) && diag.code === diagnostic.data.code)) {
                    const fixAllCodeAction = reg.getFixAllCodeAction(params);
                    if (fixAllCodeAction) {
                        codeActions === null || codeActions === void 0 ? void 0 : codeActions.push(fixAllCodeAction);
                    }
                }
                return codeActions;
            }));
        });
        results.push(...treeUtils_1.flatMap(Array.from(CodeActionProvider.refactorRegistrations.values()), (registration) => registration.getAvailableActions(params)));
        return [...results, ...make, ...elmDiagnostics];
    }
    onCodeActionResolve(codeAction, program, sourceFile) {
        var _a;
        const result = (_a = CodeActionProvider.refactorRegistrations
            .get(codeAction.data.refactorName)) === null || _a === void 0 ? void 0 : _a.getEditsForAction({
            textDocument: {
                uri: codeAction.data.uri,
            },
            context: { diagnostics: [] },
            range: codeAction.data.range,
            program,
            sourceFile,
        }, codeAction.data.actionName);
        if (result === null || result === void 0 ? void 0 : result.edits) {
            codeAction.edit = { changes: { [codeAction.data.uri]: result.edits } };
        }
        if (result === null || result === void 0 ? void 0 : result.renamePosition) {
            codeAction.data.renamePosition = result.renamePosition;
        }
        return codeAction;
    }
    addDiagnosticToCodeAction(codeAction, diagnostic) {
        codeAction.diagnostics = [diagnostic];
        return codeAction;
    }
}
exports.CodeActionProvider = CodeActionProvider;
CodeActionProvider.errorCodeToRegistrationMap = new multiMap_1.MultiMap();
CodeActionProvider.refactorRegistrations = new Map();
//# sourceMappingURL=codeActionProvider.js.map