"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getSpaces = exports.RefactorEditUtils = void 0;
const vscode_languageserver_1 = require("vscode-languageserver");
const treeUtils_1 = require("./treeUtils");
class RefactorEditUtils {
    static findLineNumberAfterCurrentFunction(nodeAtPosition) {
        var _a;
        if (!nodeAtPosition.parent) {
            return undefined;
        }
        if (((_a = nodeAtPosition.parent) === null || _a === void 0 ? void 0 : _a.type) === "file") {
            return nodeAtPosition.endPosition.row + 1;
        }
        return this.findLineNumberAfterCurrentFunction(nodeAtPosition.parent);
    }
    static unexposedValueInModule(tree, valueName) {
        const exposedNodes = treeUtils_1.TreeUtils.getModuleExposingListNodes(tree);
        if (exposedNodes.length <= 1) {
            // We can't remove the last exposed one and removing the whole module annotation would just lead to elm-format readding it
            return undefined;
        }
        else {
            return this.removeValueFromExposingList(exposedNodes, valueName);
        }
    }
    static createTopLevelFunction(insertLineNumber, valueName, typeString, args, content) {
        return this.createFunction(insertLineNumber, valueName, typeString, args, content);
    }
    static createFunction(insertLineNumber, valueName, typeString, args, content, contentIndendation = 0, targetIndendation = 0) {
        const hasArity0 = args === 0 || (Array.isArray(args) && args.length === 0);
        const argList = typeof args === "number" ? this.argListFromArity(args) : args.join(" ");
        const bodyTargetIndendation = targetIndendation + 4;
        const spaces = getSpaces(targetIndendation);
        const bodySpaces = getSpaces(bodyTargetIndendation);
        const diffIndentation = bodyTargetIndendation - contentIndendation;
        if (diffIndentation > 0) {
            const diffSpaces = getSpaces(diffIndentation);
            content = content.split("\n").join(`\n${diffSpaces}`);
        }
        else if (diffIndentation < 0) {
            content = content
                .split("\n")
                .map((line) => {
                let i = 0;
                while (i < Math.abs(diffIndentation) && line[i] === " ") {
                    i++;
                }
                return line.slice(i);
            })
                .join("\n");
        }
        if (hasArity0) {
            return vscode_languageserver_1.TextEdit.insert(vscode_languageserver_1.Position.create(insertLineNumber, 0), `\n\n${typeString ? `${spaces}${valueName}` + " : " + typeString + "\n" : ""}${spaces}${valueName} =\n${bodySpaces}${content}\n`);
        }
        else {
            return vscode_languageserver_1.TextEdit.insert(vscode_languageserver_1.Position.create(insertLineNumber, 0), `\n\n${typeString ? `${spaces}${valueName}` + " : " + typeString + "\n" : ""}${spaces}${valueName} ${argList} =\n${bodySpaces}${content}\n`);
        }
    }
    static argListFromArity(arity) {
        return [...Array(arity).keys()].map((a) => `arg${a + 1}`).join(" ");
    }
    static exposeValueInModule(tree, valueName) {
        const exposedNodes = treeUtils_1.TreeUtils.getModuleExposingListNodes(tree);
        if (exposedNodes.length > 0) {
            const lastExposedNode = exposedNodes[exposedNodes.length - 1];
            if (lastExposedNode) {
                return vscode_languageserver_1.TextEdit.insert(vscode_languageserver_1.Position.create(lastExposedNode.endPosition.row, lastExposedNode.endPosition.column), `, ${valueName}`);
            }
        }
    }
    static removeValueFromImport(tree, moduleName, valueName) {
        const importClause = treeUtils_1.TreeUtils.findImportClauseByName(tree, moduleName);
        if (importClause) {
            const exposedValuesAndTypes = [
                ...importClause.descendantsOfType(["exposed_value", "exposed_type"]),
            ];
            if (exposedValuesAndTypes.length === 1 &&
                exposedValuesAndTypes[0].text === valueName) {
                // Remove the entire exposing list if it was the only one
                const exposingList = treeUtils_1.TreeUtils.findFirstNamedChildOfType("exposing_list", importClause);
                if (exposingList) {
                    return vscode_languageserver_1.TextEdit.del(vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(exposingList.startPosition.row, exposingList.startPosition.column), vscode_languageserver_1.Position.create(exposingList.endPosition.row, exposingList.endPosition.column)));
                }
            }
            else {
                return this.removeValueFromExposingList(exposedValuesAndTypes, valueName);
            }
        }
    }
    static addImport(tree, moduleName, valueName) {
        var _a, _b;
        const lastImportNode = (_a = treeUtils_1.TreeUtils.getLastImportNode(tree)) !== null && _a !== void 0 ? _a : (_b = treeUtils_1.TreeUtils.getModuleNameNode(tree)) === null || _b === void 0 ? void 0 : _b.parent;
        return vscode_languageserver_1.TextEdit.insert(vscode_languageserver_1.Position.create((lastImportNode === null || lastImportNode === void 0 ? void 0 : lastImportNode.endPosition.row) ? (lastImportNode === null || lastImportNode === void 0 ? void 0 : lastImportNode.endPosition.row) + 1
            : 1, 0), valueName
            ? `import ${moduleName} exposing (${valueName})\n`
            : `import ${moduleName}\n`);
    }
    static changeQualifiedReferenceModule(node, newModuleName) {
        if (node.parent && node.parent.type === "value_qid") {
            const moduleNode = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", node.parent);
            if (moduleNode) {
                return vscode_languageserver_1.TextEdit.replace(vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(moduleNode.startPosition.row, moduleNode.startPosition.column), vscode_languageserver_1.Position.create(moduleNode.endPosition.row, moduleNode.endPosition.column)), newModuleName);
            }
        }
    }
    static removeQualifiedReference(node) {
        if (node.parent && node.parent.type === "value_qid") {
            const moduleNode = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", node.parent);
            if (moduleNode) {
                return vscode_languageserver_1.TextEdit.del(vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(moduleNode.startPosition.row, moduleNode.startPosition.column), vscode_languageserver_1.Position.create(moduleNode.endPosition.row, moduleNode.endPosition.column + 1)));
            }
        }
    }
    static addImports(tree, importData) {
        var _a, _b;
        const lastImportNode = (_a = treeUtils_1.TreeUtils.getLastImportNode(tree)) !== null && _a !== void 0 ? _a : (_b = treeUtils_1.TreeUtils.getModuleNameNode(tree)) === null || _b === void 0 ? void 0 : _b.parent;
        const imports = importData
            .filter((data, i, array) => array.findIndex((d) => d.moduleName === data.moduleName &&
            d.valueName === data.valueName) === i)
            .map((data) => data.valueName
            ? `import ${data.moduleName} exposing (${data.valueName})`
            : `import ${data.moduleName}`)
            .join("\n")
            .concat("\n");
        return vscode_languageserver_1.TextEdit.insert(vscode_languageserver_1.Position.create((lastImportNode === null || lastImportNode === void 0 ? void 0 : lastImportNode.endPosition.row) ? (lastImportNode === null || lastImportNode === void 0 ? void 0 : lastImportNode.endPosition.row) + 1
            : 1, 0), imports);
    }
    static addModuleDeclaration(moduleName) {
        return vscode_languageserver_1.TextEdit.insert(vscode_languageserver_1.Position.create(0, 0), `module ${moduleName} exposing (..)`);
    }
    static renameModuleDeclaration(tree, newModuleName) {
        const moduleNameNode = treeUtils_1.TreeUtils.getModuleNameNode(tree);
        if (moduleNameNode) {
            return vscode_languageserver_1.TextEdit.replace(vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(moduleNameNode.startPosition.row, moduleNameNode.startPosition.column), vscode_languageserver_1.Position.create(moduleNameNode.endPosition.row, moduleNameNode.endPosition.column)), newModuleName);
        }
    }
    static removeRecordPatternValue(pattern) {
        var _a, _b, _c;
        let startPosition = pattern.startPosition;
        let endPosition = pattern.endPosition;
        if (((_a = pattern.previousNamedSibling) === null || _a === void 0 ? void 0 : _a.text) === ",") {
            startPosition = pattern.previousNamedSibling.startPosition;
        }
        if (((_b = pattern.previousNamedSibling) === null || _b === void 0 ? void 0 : _b.text) !== "," &&
            ((_c = pattern.nextNamedSibling) === null || _c === void 0 ? void 0 : _c.text) === ",") {
            endPosition = pattern.nextNamedSibling.endPosition;
        }
        return vscode_languageserver_1.TextEdit.del(vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(startPosition.row, startPosition.column), vscode_languageserver_1.Position.create(endPosition.row, endPosition.column)));
    }
    static removeValueFromExposingList(exposedNodes, valueName) {
        var _a, _b, _c;
        const exposedNode = exposedNodes.find((node) => node.text === valueName || node.text === `${valueName}(..)`);
        if (exposedNode) {
            let startPosition = exposedNode.startPosition;
            let endPosition = exposedNode.endPosition;
            if (((_a = exposedNode.previousNamedSibling) === null || _a === void 0 ? void 0 : _a.text) === ",") {
                startPosition = exposedNode.previousNamedSibling.startPosition;
            }
            if (((_b = exposedNode.previousNamedSibling) === null || _b === void 0 ? void 0 : _b.text) !== "," &&
                ((_c = exposedNode.nextNamedSibling) === null || _c === void 0 ? void 0 : _c.text) === ",") {
                endPosition = exposedNode.nextNamedSibling.endPosition;
            }
            return vscode_languageserver_1.TextEdit.del(vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(startPosition.row, startPosition.column), vscode_languageserver_1.Position.create(endPosition.row, endPosition.column)));
        }
    }
}
exports.RefactorEditUtils = RefactorEditUtils;
function getSpaces(n) {
    return Array(n + 1)
        .map(() => "")
        .join(" ");
}
exports.getSpaces = getSpaces;
//# sourceMappingURL=refactorEditUtils.js.map