import { Position } from "vscode-languageserver";
import { SyntaxNode, Tree } from "web-tree-sitter";
import { ITreeContainer } from "../forest";
import { TRecord, Type } from "./types/typeInference";
import { IElmWorkspace } from "../elmWorkspace";
import { Range } from "vscode-languageserver-textdocument";
export declare type NodeType = "Function" | "FunctionParameter" | "TypeAlias" | "Type" | "Operator" | "Module" | "CasePattern" | "AnonymousFunctionParameter" | "UnionConstructor" | "FieldType" | "TypeVariable" | "Port" | "Import";
export interface IExposed {
    name: string;
    syntaxNode: SyntaxNode;
    type: NodeType;
    exposedUnionConstructors?: {
        name: string;
        syntaxNode: SyntaxNode;
    }[];
}
export declare type IExposing = Map<string, IExposed>;
export declare function flatMap<T, U>(array: T[] | undefined, callback: (value: T, index: number, array: T[]) => U[] | undefined): U[];
export declare class TreeUtils {
    static getModuleNameNode(tree: Tree): SyntaxNode | undefined;
    static getModuleExposingListNodes(tree: Tree): SyntaxNode[];
    static findFirstNamedChildOfType(type: string, node: SyntaxNode): SyntaxNode | undefined;
    static findAllNamedChildrenOfType(type: string | string[], node: SyntaxNode): SyntaxNode[] | undefined;
    static findExposedFunctionNode(node: SyntaxNode, functionName: string): SyntaxNode | undefined;
    static isExposedFunctionOrPort(tree: Tree, functionName: string): boolean;
    static findExposedTypeOrTypeAliasNode(node: SyntaxNode, typeName: string): SyntaxNode | undefined;
    static isExposedTypeOrTypeAlias(tree: Tree, typeName: string): boolean;
    static findUnionConstructor(tree: Tree, unionConstructorName: string): SyntaxNode | undefined;
    static findUnionConstructorCalls(tree: Tree, unionConstructorName: string): SyntaxNode[] | undefined;
    static findFunction(syntaxNode: SyntaxNode, functionName: string, onlySearchTopLevel?: boolean): SyntaxNode | undefined;
    static findPort(tree: Tree, portName: string): SyntaxNode | undefined;
    static findOperator(treeContainer: ITreeContainer, operatorName: string): SyntaxNode | undefined;
    static findTypeDeclaration(tree: Tree, typeName: string): SyntaxNode | undefined;
    static findModuleDeclaration(tree: Tree): SyntaxNode | undefined;
    static findTypeAliasDeclaration(tree: Tree, typeAliasName: string): SyntaxNode | undefined;
    static findAllTopLevelFunctionDeclarations(tree: Tree): SyntaxNode[] | undefined;
    static findAllTopLevelFunctionDeclarationsWithoutTypeAnnotation(tree: Tree): SyntaxNode[] | undefined;
    static findAllTypeOrTypeAliasCalls(tree: Tree): SyntaxNode[] | undefined;
    static getFunctionNameNodeFromDefinition(node: SyntaxNode): SyntaxNode | undefined;
    static getTypeOrTypeAliasOrPortNameNodeFromDefinition(node: SyntaxNode): SyntaxNode | undefined;
    static findTypeOrTypeAliasCalls(tree: Tree, typeOrTypeAliasName: string): SyntaxNode[] | undefined;
    static findAllTypeDeclarations(tree: Tree): SyntaxNode[] | undefined;
    static findAllTypeAliasDeclarations(tree: Tree): SyntaxNode[] | undefined;
    static findTypeAliasTypeVariable(nodeAtPosition: SyntaxNode, nodeAtPositionText: string): SyntaxNode | undefined;
    /**
     * @deprecated Should not be used due to performance. Use bindings instead
     */
    static findImportClauseByName(tree: Tree, moduleName: string): SyntaxNode | undefined;
    static getTypeOrTypeAliasOfFunctionParameter(node: SyntaxNode | undefined): SyntaxNode | undefined;
    static getReturnTypeOrTypeAliasOfFunctionDefinition(node: SyntaxNode | undefined): SyntaxNode | undefined;
    static getRecordTypeOfFunctionRecordParameter(node: SyntaxNode | undefined, elmWorkspace: IElmWorkspace): TRecord | undefined;
    static getTypeAliasOfRecordField(node: SyntaxNode | undefined, treeContainer: ITreeContainer, elmWorkspace: IElmWorkspace): {
        node: SyntaxNode;
        uri: string;
    } | undefined;
    static getTypeAliasOfCase(type: SyntaxNode | undefined, treeContainer: ITreeContainer, elmWorkspace: IElmWorkspace): {
        node: SyntaxNode;
        uri: string;
    } | undefined;
    static getTypeAliasOfRecord(node: SyntaxNode | undefined, treeContainer: ITreeContainer, elmWorkspace: IElmWorkspace): {
        node: SyntaxNode;
        uri: string;
    } | undefined;
    static getAllFieldsFromTypeAlias(node: SyntaxNode | undefined): {
        field: string;
        type: string;
    }[] | undefined;
    static descendantsOfType(node: SyntaxNode, type: string): SyntaxNode[];
    static getNamedDescendantForPosition(node: SyntaxNode, position: Position): SyntaxNode;
    static getNamedDescendantForRange(sourceFile: ITreeContainer, range: Range): SyntaxNode;
    static findPreviousNode(node: SyntaxNode, position: Position): SyntaxNode | undefined;
    static getNamedDescendantForLineBeforePosition(node: SyntaxNode, position: Position): SyntaxNode;
    static getNamedDescendantForLineAfterPosition(node: SyntaxNode, position: Position): SyntaxNode;
    static findParentOfType(typeToLookFor: string, node: SyntaxNode, topLevel?: boolean): SyntaxNode | undefined;
    static getLastImportNode(tree: Tree): SyntaxNode | undefined;
    static isReferenceFullyQualified(node: SyntaxNode): boolean;
    static getTypeAnnotation(valueDeclaration?: SyntaxNode): SyntaxNode | undefined;
    static getValueDeclaration(typeAnnotation?: SyntaxNode): SyntaxNode | undefined;
    /**
     * This gets a list of all ancestors of a type
     * in order from the closest declaration up to the top level declaration
     */
    static getAllAncestorsOfType(type: string, node: SyntaxNode): SyntaxNode[];
    /**
     * @deprecated Should not be used due to performance. Use bindings instead
     */
    static findAllImportClauseNodes(tree: Tree): SyntaxNode[] | undefined;
    static isIdentifier(node: SyntaxNode): boolean;
    static isImport(node: SyntaxNode): boolean;
    static nextNode(node: SyntaxNode): SyntaxNode | undefined;
    static findFieldReference(type: Type, fieldName: string): {
        node: SyntaxNode;
        uri: string;
        nodeType: NodeType;
    } | undefined;
    static findImportAliasOfModule(moduleName: string, tree: Tree): string | undefined;
}
