"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.bindTreeContainer = void 0;
const multiMap_1 = require("../multiMap");
const treeUtils_1 = require("../treeUtils");
const utils_1 = require("../utils");
const syntaxNodeMap_1 = require("./syntaxNodeMap");
const diagnostics_1 = require("./diagnostics");
function createSymbolMap() {
    return new multiMap_1.MultiMap();
}
function bindTreeContainer(treeContainer) {
    if (treeContainer.symbolLinks) {
        return;
    }
    const symbolLinks = new syntaxNodeMap_1.SyntaxNodeMap();
    const nonShadowableNames = new Set();
    let container;
    let parent;
    const treeCursor = treeContainer.tree.walk();
    bind();
    treeContainer.symbolLinks = symbolLinks;
    treeContainer.nonShadowableNames = nonShadowableNames;
    // Bind exposing must happen after symbolLinks is bound
    bindExposing();
    function forEachChild(func) {
        if (treeCursor.gotoFirstChild()) {
            func();
            while (treeCursor.gotoNextSibling()) {
                func();
            }
            treeCursor.gotoParent();
        }
    }
    function bind() {
        const node = treeCursor.currentNode();
        switch (node.type) {
            case "file":
            case "let_in_expr":
            case "anonymous_function_expr":
            case "case_of_branch":
                bindContainer(node);
                break;
            case "value_declaration":
                bindValueDeclaration(node);
                break;
            case "function_declaration_left":
                bindFunctionDeclarationLeft(node);
                break;
            case "type_declaration":
                bindTypeDeclaration(node);
                break;
            case "type_alias_declaration":
                bindTypeAliasDeclaration(node);
                break;
            case "lower_type_name":
                bindLowerTypeName(node);
                break;
            case "port_annotation":
                bindPortAnnotation(node);
                break;
            case "infix_declaration":
                bindInfixDeclaration(node);
                break;
            case "pattern":
                bindPattern(node);
                break;
            case "import_clause":
                bindImportClause(node);
                break;
            case "ERROR":
                treeContainer.parseDiagnostics.push(diagnostics_1.error(node, diagnostics_1.Diagnostics.Parsing));
                break;
            default:
                forEachChild(bind);
        }
    }
    function bindContainer(node) {
        const savedContainer = container;
        const savedParent = parent;
        container = createSymbolMap();
        parent = node;
        symbolLinks.set(node, container);
        forEachChild(bind);
        container = savedContainer;
        parent = savedParent;
    }
    function bindValueDeclaration(node) {
        var _a;
        // Bind the function name
        const functionDeclarationLeft = node.childForFieldName("functionDeclarationLeft");
        if (functionDeclarationLeft && functionDeclarationLeft.firstChild) {
            const functionName = functionDeclarationLeft.firstChild.text;
            container.set(functionName, {
                node: functionDeclarationLeft,
                type: "Function",
            });
            // Add to nonShadowableNames if it is top level
            if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "file") {
                nonShadowableNames.add(functionDeclarationLeft.firstChild.text);
            }
        }
        else {
            // If there is a pattern, bind it to the parent container
            const pattern = node.childForFieldName("pattern");
            if (pattern) {
                pattern.descendantsOfType("lower_pattern").forEach((lowerPattern) => {
                    container.set(lowerPattern.text, {
                        node: lowerPattern,
                        type: "Function",
                    });
                });
            }
        }
        // Bind the rest of the container
        bindContainer(node);
    }
    function bindFunctionDeclarationLeft(node) {
        node.descendantsOfType("lower_pattern").forEach((lowerPattern) => {
            container.set(lowerPattern.text, {
                node: lowerPattern,
                type: "FunctionParameter",
            });
        });
    }
    function bindTypeDeclaration(node) {
        var _a;
        const name = node.childForFieldName("name");
        if (name) {
            container.set(name.text, { node, type: "Type" });
        }
        // Union variants get bound to the parent container
        (_a = treeUtils_1.TreeUtils.findAllNamedChildrenOfType("union_variant", node)) === null || _a === void 0 ? void 0 : _a.forEach((unionVariant) => {
            const name = unionVariant.childForFieldName("name");
            if (name) {
                container.set(name.text, {
                    node: unionVariant,
                    type: "UnionConstructor",
                });
            }
        });
        // Bind type variables
        bindContainer(node);
    }
    function bindTypeAliasDeclaration(node) {
        const name = node.childForFieldName("name");
        if (name) {
            container.set(name.text, { node, type: "TypeAlias" });
        }
        // Bind type variables
        bindContainer(node);
    }
    function bindLowerTypeName(node) {
        container.set(node.text, { node, type: "TypeVariable" });
    }
    function bindPortAnnotation(node) {
        // TODO: Use field
        const name = treeUtils_1.TreeUtils.findFirstNamedChildOfType("lower_case_identifier", node);
        if (name) {
            container.set(name.text, { node, type: "Port" });
        }
    }
    function bindInfixDeclaration(node) {
        const operator = node.childForFieldName("operator");
        const name = node.lastNamedChild;
        if (operator && name) {
            container.set(operator.text, { node, type: "Operator" });
            container.set(name.text, { node, type: "Operator" });
        }
    }
    function bindPattern(node) {
        node.descendantsOfType("lower_pattern").forEach((lowerPattern) => {
            switch (parent.type) {
                case "anonymous_function_expr":
                    container.set(lowerPattern.text, {
                        node: lowerPattern,
                        type: "AnonymousFunctionParameter",
                    });
                    break;
                case "case_of_branch":
                    container.set(lowerPattern.text, {
                        node: lowerPattern,
                        type: "CasePattern",
                    });
                    break;
            }
        });
    }
    function bindImportClause(node) {
        const asClause = node.childForFieldName("asClause");
        let name;
        if (asClause) {
            name = asClause.childForFieldName("name");
        }
        else {
            name = node.childForFieldName("moduleName");
        }
        if (name) {
            container.set(name.text, { node, type: "Import" });
        }
    }
    function bindExposing() {
        const tree = treeContainer.tree;
        const exposed = new Map();
        const moduleDeclaration = treeUtils_1.TreeUtils.findModuleDeclaration(tree);
        if (moduleDeclaration) {
            const exposingList = moduleDeclaration.childForFieldName("exposing");
            if (exposingList) {
                const rootSymbols = symbolLinks.get(tree.rootNode);
                if (exposingList.childForFieldName("doubleDot")) {
                    rootSymbols === null || rootSymbols === void 0 ? void 0 : rootSymbols.forEach((symbol) => {
                        var _a, _b, _c;
                        switch (symbol.type) {
                            case "Function":
                                {
                                    if (symbol.node.firstNamedChild) {
                                        const functionName = symbol.node.firstNamedChild.text;
                                        exposed.set(functionName, {
                                            name: functionName,
                                            syntaxNode: symbol.node,
                                            type: "Function",
                                        });
                                    }
                                }
                                break;
                            case "TypeAlias":
                                {
                                    const name = symbol.node.childForFieldName("name");
                                    if (name) {
                                        exposed.set(name.text, {
                                            name: name.text,
                                            syntaxNode: symbol.node,
                                            type: "TypeAlias",
                                        });
                                    }
                                }
                                break;
                            case "Type":
                                {
                                    const unionConstructors = (_b = (_a = treeUtils_1.TreeUtils.findAllNamedChildrenOfType("union_variant", symbol.node)) === null || _a === void 0 ? void 0 : _a.map((variant) => {
                                        const name = variant.childForFieldName("name");
                                        if (name && name.parent) {
                                            return {
                                                name: name.text,
                                                syntaxNode: variant,
                                            };
                                        }
                                    }).filter(utils_1.Utils.notUndefined.bind(bindExposing))) !== null && _b !== void 0 ? _b : [];
                                    const typeDeclarationName = symbol.node.childForFieldName("name");
                                    if (typeDeclarationName) {
                                        exposed.set(typeDeclarationName.text, {
                                            name: typeDeclarationName.text,
                                            syntaxNode: symbol.node,
                                            type: "Type",
                                            exposedUnionConstructors: unionConstructors,
                                        });
                                    }
                                }
                                break;
                            case "Port":
                                {
                                    const name = (_c = symbol.node.childForFieldName("name")) === null || _c === void 0 ? void 0 : _c.text;
                                    if (name) {
                                        exposed.set(name, {
                                            name,
                                            syntaxNode: symbol.node,
                                            type: "Port",
                                        });
                                    }
                                }
                                break;
                        }
                    });
                }
                else {
                    const exposedOperators = treeUtils_1.TreeUtils.descendantsOfType(exposingList, "operator_identifier");
                    for (const value of exposedOperators) {
                        const functionNode = treeUtils_1.TreeUtils.findOperator(treeContainer, value.text);
                        if (functionNode) {
                            exposed.set(value.text, {
                                name: value.text,
                                syntaxNode: functionNode,
                                type: "Operator",
                                exposedUnionConstructors: undefined,
                            });
                        }
                    }
                    treeUtils_1.TreeUtils.descendantsOfType(exposingList, "exposed_value")
                        .map((a) => a.text)
                        .forEach((exposedValue) => {
                        const symbol = rootSymbols === null || rootSymbols === void 0 ? void 0 : rootSymbols.get(exposedValue);
                        if (symbol) {
                            exposed.set(exposedValue, {
                                name: exposedValue,
                                syntaxNode: symbol.node,
                                type: symbol.type,
                            });
                        }
                    });
                    const exposedTypes = treeUtils_1.TreeUtils.descendantsOfType(exposingList, "exposed_type");
                    for (const value of exposedTypes) {
                        const doubleDot = treeUtils_1.TreeUtils.descendantsOfType(value, "double_dot");
                        if (doubleDot.length > 0) {
                            const name = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", value);
                            if (name) {
                                const typeDeclaration = rootSymbols === null || rootSymbols === void 0 ? void 0 : rootSymbols.get(name.text, (s) => s.type === "Type");
                                if (typeDeclaration) {
                                    const unionConstructors = treeUtils_1.TreeUtils.descendantsOfType(typeDeclaration.node, "union_variant")
                                        .map((variant) => {
                                        const unionConstructorName = variant.childForFieldName("name");
                                        if (unionConstructorName && unionConstructorName.parent) {
                                            return {
                                                name: unionConstructorName.text,
                                                syntaxNode: variant,
                                            };
                                        }
                                    })
                                        .filter(utils_1.Utils.notUndefined.bind(bindExposing));
                                    exposed.set(name.text, {
                                        name: name.text,
                                        syntaxNode: typeDeclaration.node,
                                        type: "Type",
                                        exposedUnionConstructors: unionConstructors,
                                    });
                                }
                            }
                        }
                        else {
                            const typeNode = rootSymbols === null || rootSymbols === void 0 ? void 0 : rootSymbols.get(value.text, (symbol) => symbol.type === "Type" || symbol.type === "TypeAlias");
                            if (typeNode) {
                                exposed.set(value.text, {
                                    name: value.text,
                                    syntaxNode: typeNode.node,
                                    type: typeNode.type,
                                });
                            }
                        }
                    }
                }
            }
        }
        treeContainer.exposing = exposed;
    }
}
exports.bindTreeContainer = bindTreeContainer;
//# sourceMappingURL=binder.js.map