"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createTypeChecker = exports.resetBindTime = exports.bindTime = void 0;
const treeUtils_1 = require("../treeUtils");
const expressionTree_1 = require("./expressionTree");
const tsyringe_1 = require("tsyringe");
const typeInference_1 = require("./typeInference");
const imports_1 = require("../../imports");
const typeRenderer_1 = require("./typeRenderer");
const perf_hooks_1 = require("perf_hooks");
const binder_1 = require("./binder");
const sequence_1 = require("../sequence");
const utils_1 = require("../utils");
const typeExpression_1 = require("./typeExpression");
const diagnostics_1 = require("./diagnostics");
exports.bindTime = 0;
function resetBindTime() {
    exports.bindTime = 0;
}
exports.resetBindTime = resetBindTime;
class DiagnosticsCollection extends Map {
    add(diagnostic) {
        let diagnostics = super.get(diagnostic.uri);
        if (!diagnostics) {
            diagnostics = [];
            super.set(diagnostic.uri, diagnostics);
        }
        diagnostics.push(diagnostic);
    }
    get(uri) {
        var _a;
        return (_a = super.get(uri)) !== null && _a !== void 0 ? _a : [];
    }
}
function createTypeChecker(workspace) {
    const forest = workspace.getForest();
    const imports = new Map();
    const diagnostics = new DiagnosticsCollection();
    const suggestionDiagnostics = new DiagnosticsCollection();
    let cancellationToken;
    const checkedNodes = new Set();
    const start = perf_hooks_1.performance.now();
    forest.treeMap.forEach((treeContainer) => {
        binder_1.bindTreeContainer(treeContainer);
    });
    exports.bindTime = perf_hooks_1.performance.now() - start;
    const typeChecker = {
        findType,
        findDefinition,
        findDefinitionShallow,
        getAllImports,
        getQualifierForName,
        typeToString,
        getDiagnostics,
        getDiagnosticsAsync,
        getSuggestionDiagnostics,
        findImportModuleNameNode,
    };
    return typeChecker;
    function findType(node) {
        var _a, _b, _c, _d, _e;
        try {
            const declaration = expressionTree_1.mapSyntaxNodeToExpression(treeUtils_1.TreeUtils.findParentOfType("value_declaration", node, true));
            const uri = node.tree.uri;
            const findTypeOrParentType = (expr, inferenceResult) => {
                var _a, _b, _c;
                const found = expr
                    ? inferenceResult.expressionTypes.get(expr)
                    : undefined;
                if (found) {
                    return found;
                }
                // Check if the parent is the same text and position
                if (expr &&
                    expr.text === ((_a = expr.parent) === null || _a === void 0 ? void 0 : _a.text) &&
                    expr.startIndex === ((_b = expr.parent) === null || _b === void 0 ? void 0 : _b.startIndex) &&
                    expr.endIndex === ((_c = expr.parent) === null || _c === void 0 ? void 0 : _c.endIndex)) {
                    return findTypeOrParentType(expr.parent, inferenceResult);
                }
            };
            if (declaration && declaration.nodeType === "ValueDeclaration") {
                const inferenceResult = typeInference_1.InferenceScope.valueDeclarationInference(declaration, uri, workspace, new Set());
                if ((node === null || node === void 0 ? void 0 : node.type) === "function_declaration_left") {
                    const declaration = treeUtils_1.TreeUtils.findParentOfType("value_declaration", node);
                    if (declaration) {
                        return ((_a = inferenceResult.expressionTypes.get(declaration)) !== null && _a !== void 0 ? _a : inferenceResult.type);
                    }
                    else {
                        return typeInference_1.TUnknown;
                    }
                }
                else if (node.type === "value_declaration") {
                    return ((_b = inferenceResult.expressionTypes.get(node)) !== null && _b !== void 0 ? _b : inferenceResult.type);
                }
                return (_c = findTypeOrParentType(node, inferenceResult)) !== null && _c !== void 0 ? _c : typeInference_1.TUnknown;
            }
            const typeAliasDeclaration = expressionTree_1.mapSyntaxNodeToExpression(treeUtils_1.TreeUtils.findParentOfType("type_alias_declaration", node));
            if (typeAliasDeclaration &&
                typeAliasDeclaration.nodeType === "TypeAliasDeclaration") {
                const inferenceResult = typeExpression_1.TypeExpression.typeAliasDeclarationInference(typeAliasDeclaration, workspace);
                if (node.type === "type_alias_declaration") {
                    return inferenceResult.type;
                }
                return (_d = findTypeOrParentType(node, inferenceResult)) !== null && _d !== void 0 ? _d : typeInference_1.TUnknown;
            }
            const unionVariant = node.type === "union_variant"
                ? expressionTree_1.mapSyntaxNodeToExpression(node)
                : undefined;
            if (unionVariant && unionVariant.nodeType === "UnionVariant") {
                return typeExpression_1.TypeExpression.unionVariantInference(unionVariant, workspace)
                    .type;
            }
            const typeDeclaration = expressionTree_1.mapSyntaxNodeToExpression(treeUtils_1.TreeUtils.findParentOfType("type_declaration", node));
            if (typeDeclaration && typeDeclaration.nodeType === "TypeDeclaration") {
                const inferenceResult = typeExpression_1.TypeExpression.typeDeclarationInference(typeDeclaration, workspace);
                return (_e = findTypeOrParentType(node, inferenceResult)) !== null && _e !== void 0 ? _e : typeInference_1.TUnknown;
            }
            const portAnnotation = expressionTree_1.mapSyntaxNodeToExpression(node);
            if (portAnnotation && portAnnotation.nodeType === "PortAnnotation") {
                return typeExpression_1.TypeExpression.portAnnotationInference(portAnnotation, workspace)
                    .type;
            }
            return typeInference_1.TUnknown;
        }
        catch (error) {
            const connection = tsyringe_1.container.resolve("Connection");
            connection.console.warn(`Error while trying to infer a type. ${error}`);
            return typeInference_1.TUnknown;
        }
    }
    function getDiagnostics(treeContainer, token) {
        try {
            cancellationToken = token;
            checkNode(treeContainer.tree.rootNode);
            return diagnostics.get(treeContainer.uri);
        }
        finally {
            cancellationToken = undefined;
        }
    }
    function getDiagnosticsAsync(treeContainer, token, cancelCallback) {
        cancellationToken = token;
        return new Promise((resolve, reject) => {
            const children = treeContainer.tree.rootNode.children;
            let index = 0;
            const goNext = () => {
                index++;
                if (children.length > index) {
                    setImmediate(checkOne);
                }
                else {
                    cancellationToken = undefined;
                    resolve(diagnostics.get(treeContainer.uri));
                }
            };
            const checkOne = () => {
                if (cancelCallback && cancelCallback()) {
                    reject();
                    return;
                }
                try {
                    checkNode(children[index]);
                }
                catch (_a) {
                    cancellationToken = undefined;
                    reject();
                    return;
                }
                goNext();
            };
            checkOne();
        });
    }
    function getSuggestionDiagnostics(treeContainer, token) {
        try {
            cancellationToken = token;
            checkNode(treeContainer.tree.rootNode);
            return suggestionDiagnostics.get(treeContainer.uri);
        }
        finally {
            cancellationToken = undefined;
        }
    }
    function getAllImports(treeContainer) {
        const cached = imports.get(treeContainer.uri);
        if (cached) {
            return cached;
        }
        const allImports = imports_1.Imports.getImports(treeContainer, forest);
        imports.set(treeContainer.uri, allImports);
        return allImports;
    }
    function findImport(treeContainer, name, filter) {
        const possibleImport = getAllImports(treeContainer).get(name, filter);
        if (possibleImport) {
            return {
                node: possibleImport.node,
                nodeType: possibleImport.type,
                uri: possibleImport.fromUri,
            };
        }
    }
    function findImportOfType(treeContainer, name, type) {
        return findImport(treeContainer, name, (imp) => imp.type === type);
    }
    function findDefinition(nodeAtPosition, treeContainer) {
        const definition = findDefinitionShallow(nodeAtPosition, treeContainer);
        if ((definition === null || definition === void 0 ? void 0 : definition.node.type) === "lower_pattern" &&
            definition.node.firstNamedChild) {
            const innerDefinition = findDefinitionShallow(definition.node.firstNamedChild, treeContainer);
            if (innerDefinition) {
                return innerDefinition;
            }
        }
        return definition;
    }
    function findDefinitionShallow(nodeAtPosition, treeContainer) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s, _t, _u, _v, _w, _x, _y, _z, _0, _1, _2, _3, _4, _5, _6;
        const uri = treeContainer.uri;
        const nodeText = nodeAtPosition.text;
        const nodeParent = nodeAtPosition.parent;
        if (!nodeParent) {
            return;
        }
        const nodeParentType = nodeParent.type;
        const rootSymbols = (_a = treeContainer.symbolLinks) === null || _a === void 0 ? void 0 : _a.get(treeContainer.tree.rootNode);
        if (nodeParentType === "upper_case_qid" &&
            ((_b = nodeParent.previousNamedSibling) === null || _b === void 0 ? void 0 : _b.type) === "module") {
            const moduleNode = nodeParent.parent;
            if (moduleNode) {
                return {
                    node: moduleNode,
                    nodeType: "Module",
                    uri,
                };
            }
        }
        else if (nodeParentType === "upper_case_qid" &&
            ((_c = nodeParent.previousNamedSibling) === null || _c === void 0 ? void 0 : _c.type) === "import") {
            const upperCaseQid = nodeParent;
            const upperCaseQidText = upperCaseQid.text;
            return findImportOfType(treeContainer, upperCaseQidText, "Module");
        }
        else if ((nodeParentType === "exposed_value" &&
            ((_e = (_d = nodeParent.parent) === null || _d === void 0 ? void 0 : _d.parent) === null || _e === void 0 ? void 0 : _e.type) === "module_declaration") ||
            nodeParentType === "type_annotation" ||
            nodeParentType === "port_annotation") {
            const definitionNode = rootSymbols === null || rootSymbols === void 0 ? void 0 : rootSymbols.get(nodeText);
            if (definitionNode &&
                (definitionNode.type === "Function" || definitionNode.type === "Port")) {
                return {
                    node: definitionNode.node,
                    nodeType: definitionNode.type,
                    uri,
                };
            }
        }
        else if ((nodeParentType === "exposed_type" &&
            ((_g = (_f = nodeParent.parent) === null || _f === void 0 ? void 0 : _f.parent) === null || _g === void 0 ? void 0 : _g.type) === "module_declaration") ||
            ((_h = nodeAtPosition.previousNamedSibling) === null || _h === void 0 ? void 0 : _h.type) === "type" ||
            ((_j = nodeAtPosition.previousNamedSibling) === null || _j === void 0 ? void 0 : _j.type) === "alias") {
            const definitionNode = rootSymbols === null || rootSymbols === void 0 ? void 0 : rootSymbols.get(nodeText);
            if (definitionNode) {
                return {
                    node: definitionNode.node,
                    nodeType: definitionNode.type,
                    uri,
                };
            }
        }
        else if ((nodeParentType === "exposed_value" ||
            nodeParentType === "exposed_type") &&
            ((_l = (_k = nodeParent.parent) === null || _k === void 0 ? void 0 : _k.parent) === null || _l === void 0 ? void 0 : _l.type) === "import_clause") {
            return findImport(treeContainer, nodeText);
        }
        else if (nodeParentType === "union_variant") {
            const definitionNode = nodeParent;
            return {
                node: definitionNode,
                nodeType: "UnionConstructor",
                uri,
            };
        }
        else if (nodeParent && nodeParent.type === "upper_case_qid") {
            const upperCaseQid = nodeParent;
            const upperCaseQidText = upperCaseQid.text;
            // Usage is either a type or a constructor
            // A type can only be used as a type
            // A union variant can only be used as a constructor or as a pattern
            // A type alias can be used as both a type and a constructor
            const isTypeUsage = treeUtils_1.TreeUtils.findParentOfType("type_ref", upperCaseQid) ||
                ((_m = upperCaseQid.parent) === null || _m === void 0 ? void 0 : _m.type) === "exposed_type";
            const isConstructorUsage = ((_o = upperCaseQid.parent) === null || _o === void 0 ? void 0 : _o.type) === "value_expr";
            const definitionNode = rootSymbols === null || rootSymbols === void 0 ? void 0 : rootSymbols.get(upperCaseQidText, (symbol) => {
                var _a, _b;
                return isTypeUsage
                    ? symbol.type === "Type" || symbol.type === "TypeAlias"
                    : isConstructorUsage
                        ? symbol.type === "UnionConstructor" ||
                            (symbol.type === "TypeAlias" &&
                                ((_b = (_a = symbol.node.childForFieldName("typeExpression")) === null || _a === void 0 ? void 0 : _a.children[0]) === null || _b === void 0 ? void 0 : _b.type) === "record_type")
                        : symbol.type === "UnionConstructor";
            });
            if (definitionNode) {
                return {
                    node: definitionNode.node,
                    nodeType: definitionNode.type,
                    uri,
                };
            }
            let definitionFromOtherFile;
            // Make sure the next node is a dot, or else it isn't a Module
            if (((_p = treeUtils_1.TreeUtils.nextNode(nodeAtPosition)) === null || _p === void 0 ? void 0 : _p.type) === "dot") {
                const endPos = upperCaseQidText.indexOf(nodeText) + nodeText.length;
                const moduleNameOrAlias = nodeParent.text.substring(0, endPos);
                const moduleName = (_r = (_q = findImportModuleNameNode(moduleNameOrAlias, treeContainer)) === null || _q === void 0 ? void 0 : _q.text) !== null && _r !== void 0 ? _r : moduleNameOrAlias;
                const definitionFromOtherFile = findImportOfType(treeContainer, moduleName, "Module");
                if (definitionFromOtherFile) {
                    return definitionFromOtherFile;
                }
            }
            if (isTypeUsage) {
                definitionFromOtherFile = findImportOfType(treeContainer, upperCaseQidText, "Type");
                if (definitionFromOtherFile) {
                    return definitionFromOtherFile;
                }
            }
            else {
                definitionFromOtherFile = findImportOfType(treeContainer, upperCaseQidText, "UnionConstructor");
                if (definitionFromOtherFile) {
                    return definitionFromOtherFile;
                }
            }
            definitionFromOtherFile = findImportOfType(treeContainer, upperCaseQidText, "TypeAlias");
            if (definitionFromOtherFile) {
                return definitionFromOtherFile;
            }
            definitionFromOtherFile = findImportOfType(treeContainer, (_t = (_s = findImportModuleNameNode(upperCaseQidText, treeContainer)) === null || _s === void 0 ? void 0 : _s.text) !== null && _t !== void 0 ? _t : upperCaseQidText, "Module");
            if (definitionFromOtherFile) {
                return definitionFromOtherFile;
            }
        }
        else if (nodeParentType === "lower_pattern" &&
            ((_u = nodeParent.parent) === null || _u === void 0 ? void 0 : _u.type) === "record_pattern") {
            const type = findType(nodeParent.parent);
            return treeUtils_1.TreeUtils.findFieldReference(type, nodeText);
        }
        else if (nodeAtPosition.type === "lower_case_identifier" &&
            nodeParentType === "function_declaration_left" &&
            ((_v = nodeParent.parent) === null || _v === void 0 ? void 0 : _v.type) === "value_declaration") {
            // The function name should resolve to itself
            if (((_w = nodeParent.firstNamedChild) === null || _w === void 0 ? void 0 : _w.text) === nodeText) {
                return {
                    node: nodeParent.parent,
                    nodeType: "Function",
                    uri: treeContainer.uri,
                };
            }
        }
        else if (nodeParentType === "value_qid" ||
            nodeParentType === "lower_pattern" ||
            nodeParentType === "record_base_identifier") {
            let nodeAtPositionText = nodeText;
            if (nodeParentType === "value_qid") {
                nodeAtPositionText = nodeParent.text;
            }
            // Traverse the parents and find a binding
            // For operator functions, there are two bindings:
            // The infix declaration and the function
            // Never resolve to the infix declaration
            const localBinding = new sequence_1.Sequence(nodeAtPosition, (node) => { var _a; return (_a = node.parent) !== null && _a !== void 0 ? _a : undefined; })
                .map((node) => { var _a, _b; return (_b = (_a = treeContainer.symbolLinks) === null || _a === void 0 ? void 0 : _a.get(node)) === null || _b === void 0 ? void 0 : _b.get(nodeAtPositionText, (s) => s.node.type !== "infix_declaration"); })
                .find(utils_1.Utils.notUndefined.bind(findDefinition));
            if (localBinding) {
                return {
                    node: localBinding.node,
                    nodeType: localBinding.type,
                    uri: treeContainer.uri,
                };
            }
            else {
                const nodeParentText = nodeParent.text;
                // Get the full module name and handle an import alias if there is one
                if (nodeAtPosition.type === "upper_case_identifier") {
                    const moduleNameOrAlias = (_y = (_x = treeUtils_1.TreeUtils.findAllNamedChildrenOfType("upper_case_identifier", nodeParent)) === null || _x === void 0 ? void 0 : _x.map((node) => node.text).join(".")) !== null && _y !== void 0 ? _y : "";
                    const moduleName = (_0 = (_z = findImportModuleNameNode(moduleNameOrAlias, treeContainer)) === null || _z === void 0 ? void 0 : _z.text) !== null && _0 !== void 0 ? _0 : moduleNameOrAlias;
                    const moduleDefinitionFromOtherFile = findImportOfType(treeContainer, moduleName, "Module");
                    if (moduleDefinitionFromOtherFile) {
                        return moduleDefinitionFromOtherFile;
                    }
                }
                const portDefinitionFromOtherFile = findImportOfType(treeContainer, nodeParentText, "Port");
                if (portDefinitionFromOtherFile) {
                    return portDefinitionFromOtherFile;
                }
                const functionDefinitionFromOtherFile = findImportOfType(treeContainer, nodeParentText, "Function");
                if (functionDefinitionFromOtherFile) {
                    return functionDefinitionFromOtherFile;
                }
            }
        }
        else if (nodeAtPosition.type === "operator_identifier") {
            const operatorsCache = workspace.getOperatorsCache();
            const cached = operatorsCache.get(nodeText);
            if (cached) {
                return cached;
            }
            const definitionNode = treeUtils_1.TreeUtils.findOperator(treeContainer, nodeText);
            if (definitionNode) {
                const result = {
                    node: definitionNode,
                    uri,
                    nodeType: "Operator",
                };
                operatorsCache.set(nodeText, Object.freeze(result));
                return result;
            }
            else {
                const definitionFromOtherFile = findImportOfType(treeContainer, nodeText, "Operator");
                if (definitionFromOtherFile) {
                    operatorsCache.set(nodeText, Object.freeze(definitionFromOtherFile));
                    return definitionFromOtherFile;
                }
            }
        }
        else if (nodeParentType === "field_access_expr") {
            let target = nodeParent === null || nodeParent === void 0 ? void 0 : nodeParent.childForFieldName("target");
            // Adjust for parenthesis expr. Will need to change when we handle it better in inference
            if ((target === null || target === void 0 ? void 0 : target.type) === "parenthesized_expr") {
                target = target.namedChildren[1];
            }
            if (target) {
                const type = findType(target);
                return treeUtils_1.TreeUtils.findFieldReference(type, nodeText);
            }
        }
        else if (nodeAtPosition.type === "lower_case_identifier" &&
            nodeParentType === "field" &&
            ((_1 = nodeParent.parent) === null || _1 === void 0 ? void 0 : _1.type) === "record_expr") {
            const type = findType(nodeParent.parent);
            return treeUtils_1.TreeUtils.findFieldReference(type, nodeText);
        }
        else if (nodeAtPosition.type === "lower_case_identifier" &&
            nodeParentType === "field_accessor_function_expr") {
            const type = findType(nodeParent);
            if (type.nodeType === "Function") {
                const paramType = type.params[0];
                return treeUtils_1.TreeUtils.findFieldReference(paramType, nodeText);
            }
        }
        else if (nodeAtPosition.type === "lower_case_identifier" &&
            nodeParentType === "field_type") {
            return {
                node: nodeParent,
                nodeType: "FieldType",
                uri,
            };
        }
        else if (nodeAtPosition.type === "upper_case_identifier" &&
            nodeParentType === "ERROR") {
            let fullModuleName = nodeText;
            // Get fully qualified module name
            // Ex: nodeText is Attributes, we need to get Html.Attributes manually
            let currentNode = nodeAtPosition.previousNamedSibling;
            while ((currentNode === null || currentNode === void 0 ? void 0 : currentNode.type) === "dot" &&
                ((_2 = currentNode.previousNamedSibling) === null || _2 === void 0 ? void 0 : _2.type) === "upper_case_identifier") {
                fullModuleName = `${currentNode.previousNamedSibling.text}.${fullModuleName}`;
                currentNode = currentNode.previousNamedSibling.previousNamedSibling;
            }
            return findImportOfType(treeContainer, (_4 = (_3 = findImportModuleNameNode(fullModuleName, treeContainer)) === null || _3 === void 0 ? void 0 : _3.text) !== null && _4 !== void 0 ? _4 : fullModuleName, "Module");
        }
        const parentType = (_6 = (_5 = treeUtils_1.TreeUtils.findParentOfType("type_annotation", nodeAtPosition)) !== null && _5 !== void 0 ? _5 : treeUtils_1.TreeUtils.findParentOfType("type_declaration", nodeAtPosition)) !== null && _6 !== void 0 ? _6 : treeUtils_1.TreeUtils.findParentOfType("type_alias_declaration", nodeAtPosition);
        if ((parentType === null || parentType === void 0 ? void 0 : parentType.type) === "type_annotation") {
            const ancestorDeclarations = treeUtils_1.TreeUtils.getAllAncestorsOfType("value_declaration", parentType);
            const allAnnotations = [
                ...ancestorDeclarations
                    .map((n) => treeUtils_1.TreeUtils.getTypeAnnotation(n))
                    .filter((n) => !!n)
                    .reverse(),
                parentType,
            ];
            // Remove `flatMap` function when Node 10 is dropped
            const callback = (annotation) => annotation
                ? treeUtils_1.TreeUtils.descendantsOfType(annotation, "type_variable")
                : [];
            const allTypeVariables = allAnnotations.flatMap
                ? allAnnotations.flatMap(callback)
                : treeUtils_1.flatMap(allAnnotations, callback);
            const firstMatching = allTypeVariables.find((t) => t.text === nodeText);
            if (firstMatching) {
                return {
                    node: firstMatching,
                    nodeType: "TypeVariable",
                    uri,
                };
            }
        }
        else if (parentType) {
            const allTypeNames = treeUtils_1.TreeUtils.findAllNamedChildrenOfType("lower_type_name", parentType);
            const firstMatching = allTypeNames === null || allTypeNames === void 0 ? void 0 : allTypeNames.find((t) => t.text === nodeText);
            if (firstMatching) {
                return {
                    node: firstMatching,
                    nodeType: "TypeVariable",
                    uri,
                };
            }
        }
    }
    function getQualifierForName(treeContainer, module, name) {
        var _a, _b, _c;
        const found = findImport(treeContainer, name, (imp) => imp.fromModuleName === module &&
            (imp.type === "Type" ||
                imp.type === "TypeAlias" ||
                imp.type === "UnionConstructor"));
        if (found) {
            return "";
        }
        const moduleImport = (_a = findImportModuleNameNode(module, treeContainer)) === null || _a === void 0 ? void 0 : _a.parent;
        if (!moduleImport) {
            return;
        }
        const asClause = treeUtils_1.TreeUtils.findFirstNamedChildOfType("as_clause", moduleImport);
        if (asClause) {
            return `${(_c = (_b = asClause === null || asClause === void 0 ? void 0 : asClause.lastNamedChild) === null || _b === void 0 ? void 0 : _b.text) !== null && _c !== void 0 ? _c : module}.`;
        }
        return `${module}.`;
    }
    function typeToString(t, treeContainer) {
        return new typeRenderer_1.TypeRenderer(typeChecker, treeContainer).render(t);
    }
    /**
     * Get the module name node if the name is an alias
     */
    function findImportModuleNameNode(moduleNameOrAlias, treeContainer) {
        var _a, _b, _c, _d;
        return ((_d = (_c = (_b = (_a = treeContainer.symbolLinks) === null || _a === void 0 ? void 0 : _a.get(treeContainer.tree.rootNode)) === null || _b === void 0 ? void 0 : _b.get(moduleNameOrAlias, (s) => s.type === "Import")) === null || _c === void 0 ? void 0 : _c.node.childForFieldName("moduleName")) !== null && _d !== void 0 ? _d : undefined);
    }
    function checkNode(node) {
        if (checkedNodes.has(node.id)) {
            return;
        }
        cancellationToken === null || cancellationToken === void 0 ? void 0 : cancellationToken.throwIfCancellationRequested();
        switch (node.type) {
            case "file":
                node.children.forEach(checkNode);
                break;
            case "value_declaration":
                checkValueDeclaration(node);
                break;
            case "import_clause":
                checkImportClause(node);
                break;
            case "type_alias_declaration":
                checkTypeAliasDeclaration(node);
                break;
            case "type_declaration":
                checkTypeDeclaration(node);
                break;
            case "union_variant":
                checkUnionVariant(node);
                break;
            case "port_annotation":
                checkPortAnnotation(node);
                break;
        }
        checkedNodes.add(node.id);
    }
    function checkValueDeclaration(valueDeclaration) {
        var _a;
        const declaration = expressionTree_1.mapSyntaxNodeToExpression(valueDeclaration);
        const result = typeInference_1.InferenceScope.valueDeclarationInference(declaration, valueDeclaration.tree.uri, workspace, new Set(), 
        /* recursionAllowed */ false, cancellationToken);
        result.diagnostics.forEach((diagnostic) => diagnostics.add(diagnostic));
        if (!declaration.typeAnnotation) {
            const typeString = typeToString(result.type, getSourceFileOfNode(declaration));
            if (typeString &&
                typeString !== "unknown" && ((_a = declaration.firstNamedChild) === null || _a === void 0 ? void 0 : _a.firstNamedChild)) {
                suggestionDiagnostics.add(diagnostics_1.error(declaration.firstNamedChild.firstNamedChild, diagnostics_1.Diagnostics.MissingTypeAnnotation, typeString));
            }
        }
    }
    function checkImportClause(importClause) {
        const moduleNameNode = importClause.childForFieldName("moduleName");
        if (moduleNameNode) {
            const moduleName = moduleNameNode.text;
            if (!workspace.getSourceFileOfImportableModule(getSourceFileOfNode(importClause), moduleName)) {
                diagnostics.add(diagnostics_1.error(moduleNameNode, diagnostics_1.Diagnostics.ImportMissing, moduleName));
            }
        }
    }
    function checkTypeAliasDeclaration(typeAliasDeclaration) {
        typeExpression_1.TypeExpression.typeAliasDeclarationInference(expressionTree_1.mapSyntaxNodeToExpression(typeAliasDeclaration), workspace).diagnostics.forEach((diagnostic) => diagnostics.add(diagnostic));
    }
    function checkTypeDeclaration(typeDeclaration) {
        typeExpression_1.TypeExpression.typeDeclarationInference(expressionTree_1.mapSyntaxNodeToExpression(typeDeclaration), workspace).diagnostics.forEach((diagnostic) => diagnostics.add(diagnostic));
        // Need to check union variants
        typeDeclaration.children.forEach(checkNode);
    }
    function checkUnionVariant(unionVariant) {
        typeExpression_1.TypeExpression.unionVariantInference(expressionTree_1.mapSyntaxNodeToExpression(unionVariant), workspace).diagnostics.forEach((diagnostic) => diagnostics.add(diagnostic));
    }
    function checkPortAnnotation(portAnnotation) {
        typeExpression_1.TypeExpression.portAnnotationInference(expressionTree_1.mapSyntaxNodeToExpression(portAnnotation), workspace).diagnostics.forEach((diagnostic) => diagnostics.add(diagnostic));
    }
    function getSourceFileOfNode(node) {
        const treeContainer = forest.getByUri(node.tree.uri);
        if (!treeContainer) {
            throw new Error(`Can't find treeContainer by uri "${node.tree.uri}"`);
        }
        else {
            return treeContainer;
        }
    }
}
exports.createTypeChecker = createTypeChecker;
//# sourceMappingURL=typeChecker.js.map